@@ -1,4 +1,5 @@
 /*
+ *  Copyright (C) 2020-2023  The DOSBox Staging Team
  *  Copyright (C) 2002-2021  The DOSBox Team
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -24,7 +25,7 @@
 #include "mem.h"
 #endif
 
-typedef Bitu (*CallBack_Handler)(void);
+typedef Bitu (*CallBack_Handler)();
 extern CallBack_Handler CallBack_Handlers[];
 
 enum {
@@ -62,29 +63,29 @@ using callback_number_t = uint8_t;
 #define CB_SEG		0xF000
 #define CB_SOFFSET	0x1000
 
-enum {	
+enum {
 	CBRET_NONE=0,CBRET_STOP=1
 };
 
-extern uint8_t lastint;
-
 static inline RealPt CALLBACK_RealPointer(callback_number_t cb_number)
 {
-	return RealMake(CB_SEG, (uint16_t)(CB_SOFFSET + cb_number * CB_SIZE));
+	const auto offset = CB_SOFFSET + cb_number * CB_SIZE;
+	return RealMake(CB_SEG, static_cast<uint16_t>(offset));
 }
 static inline PhysPt CALLBACK_PhysPointer(callback_number_t cb_number)
 {
-	return PhysMake(CB_SEG, (uint16_t)(CB_SOFFSET + cb_number * CB_SIZE));
+	const auto offset = CB_SOFFSET + cb_number * CB_SIZE;
+	return PhysicalMake(CB_SEG, static_cast<uint16_t>(offset));
 }
 
-static inline PhysPt CALLBACK_GetBase(void) {
+static inline PhysPt CALLBACK_GetBase() {
 	return (CB_SEG << 4) + CB_SOFFSET;
 }
 
 callback_number_t CALLBACK_Allocate();
 void CALLBACK_DeAllocate(const callback_number_t in);
 
-void CALLBACK_Idle(void);
+void CALLBACK_Idle();
 
 
 void CALLBACK_RunRealInt(uint8_t intnum);
@@ -126,7 +127,7 @@ public:
 	void Uninstall();
 
 	//Only allocate a callback number
-	void Allocate(CallBack_Handler handler,const char* description=0);
+	void Allocate(CallBack_Handler handler,const char* description=nullptr);
 	uint16_t Get_callback() {
 		return m_cb_number;
 	}
