@@ -25,6 +25,24 @@
 
 #include <memory>
 
+// Project name, lower-case and without spaces
+#define DOSBOX_PROJECT_NAME "dosbox-staging"
+
+// Name of the emulator
+#define DOSBOX_NAME "DOSBox Staging"
+
+// Development team name
+#define DOSBOX_TEAM "The " DOSBOX_NAME " Team"
+
+// Copyright string
+#define DOSBOX_COPYRIGHT "(C) " DOSBOX_TEAM
+
+// Fully qualified application ID for the emulator; see
+// https://dbus.freedesktop.org/doc/dbus-specification.html#message-protocol-names
+// for more details
+#define DOSBOX_APP_ID "org.dosbox-staging.dosbox-staging"
+
+
 int sdl_main(int argc, char *argv[]);
 
 // The shutdown_requested bool is a conditional break in the parse-loop and
@@ -37,15 +55,20 @@ extern bool shutdown_requested;
         GCC_ATTRIBUTE(__format__(__printf__, 1, 2));
 
 void MSG_Add(const char*,const char*); // add messages (in UTF-8) to the language file
-const char* MSG_Get(char const *);     // get messages (adapted to current code page) from the language file
-const char* MSG_GetRaw(char const *);  // get messages (in UTF-8, without ANSI preprocessing) from the language file
+const char* MSG_Get(const char*); // get messages (adapted to current code page)
+                                  // from the language file
+const char* MSG_GetRaw(const char*); // get messages (in UTF-8, without ANSI
+                                     // preprocessing) from the language file
 bool MSG_Exists(const char*);
 
 class Section;
+class Section_prop;
 
 typedef Bitu (LoopHandler)(void);
 
-const char *DOSBOX_GetDetailedVersion() noexcept;
+const char* DOSBOX_GetVersion() noexcept;
+const char* DOSBOX_GetDetailedVersion() noexcept;
+
 double DOSBOX_GetUptime();
 
 void DOSBOX_RunMachine();
@@ -54,9 +77,11 @@ void DOSBOX_SetNormalLoop();
 
 void DOSBOX_Init(void);
 
-class Config;
-using config_ptr_t = std::unique_ptr<Config>;
-extern config_ptr_t control;
+void DOSBOX_SetMachineTypeFromConfig(Section_prop* section);
+
+int64_t DOSBOX_GetTicksDone();
+void DOSBOX_SetTicksDone(const int64_t ticks_done);
+void DOSBOX_SetTicksScheduled(const int64_t ticks_scheduled);
 
 enum SVGACards {
 	SVGA_None,
@@ -70,13 +95,15 @@ extern SVGACards svgaCard;
 extern bool mono_cga;
 
 enum MachineType {
-	// In age-order: Hercules is the oldest and VGA is the newest
-	MCH_HERC  = 1 << 0,
-	MCH_CGA   = 1 << 1,
-	MCH_TANDY = 1 << 2,
-	MCH_PCJR  = 1 << 3,
-	MCH_EGA   = 1 << 4,
-	MCH_VGA   = 1 << 5,
+	// In rough age-order: Hercules is the oldest and VGA is the newest
+	// (Tandy started out as a clone of the PCjr, so PCjr came first)
+	MCH_INVALID = 0,
+	MCH_HERC    = 1 << 0,
+	MCH_CGA     = 1 << 1,
+	MCH_TANDY   = 1 << 2,
+	MCH_PCJR    = 1 << 3,
+	MCH_EGA     = 1 << 4,
+	MCH_VGA     = 1 << 5,
 };
 
 extern MachineType machine;
@@ -87,12 +114,13 @@ inline bool is_machine(const int type) {
 #define IS_TANDY_ARCH ((machine==MCH_TANDY) || (machine==MCH_PCJR))
 #define IS_EGAVGA_ARCH ((machine==MCH_EGA) || (machine==MCH_VGA))
 #define IS_VGA_ARCH (machine==MCH_VGA)
-#define TANDY_ARCH_CASE MCH_TANDY: case MCH_PCJR
-#define EGAVGA_ARCH_CASE MCH_EGA: case MCH_VGA
-#define VGA_ARCH_CASE MCH_VGA
 
 #ifndef DOSBOX_LOGGING_H
 #include "logging.h"
 #endif // the logging system.
 
+constexpr auto DefaultMt32RomsDir   = "mt32-roms";
+constexpr auto DefaultSoundfontsDir = "soundfonts";
+constexpr auto GlShadersDir         = "glshaders";
+
 #endif /* DOSBOX_DOSBOX_H */
