@@ -4,10 +4,11 @@ set -eu
 
 # SPDX-License-Identifier: GPL-2.0-or-later
 #
-# Copyright (C) 2020-2021  Patryk Obara <patryk.obara@gmail.com>
+# Copyright (C) 2020-2023  Patryk Obara <patryk.obara@gmail.com>
 
 SCRIPT=$(basename "$0")
-readonly SCRIPT
+VERSION="1.0.2"
+readonly SCRIPT VERSION
 
 print_usage () {
 	echo "usage: $SCRIPT [-V|--verify|-d|--diff|-a|--amend] [<commit>]"
@@ -45,6 +46,7 @@ print_usage () {
 main () {
 	case ${1:-} in
 		-h|-help|--help) print_usage ;;
+		-v|--version)    echo "$SCRIPT version $VERSION" ;;
 		-d|--diff)       handle_dependencies ; shift ; format "$@" ; assert_empty_diff ;;
 		-V|--verify)     handle_dependencies ; shift ; format "$@" ; assert_empty_diff ;;
 		-a|--amend)      handle_dependencies ; shift ; format "$@" ; amend ;;
@@ -54,8 +56,9 @@ main () {
 
 handle_dependencies () {
 	assign_gnu_sed
+	assign_clang_format
 	assert_min_version git 1007010 "Use git version 1.7.10 or newer."
-	assert_min_version clang-format 15000000 "Use clang-format version 15.0.0 or newer."
+	assert_min_version "$CLANG_FORMAT" 15000000 "Use clang-format version 15.0.0 or newer."
 }
 
 SED=""
@@ -69,15 +72,39 @@ assign_gnu_sed () {
 	# No, so help the user install it and then quit.
 	else
 		echo "'sed' is not GNU and 'gsed' is not available."
-		if [[ "${OSTYPE:-}" == "darwin"* ]]; then
+		if [[ "$(uname)" == "Darwin"* ]]; then
 			echo "Install GNU sed with: brew install gnu-sed"
 		else
-			echo "Install GNU sed with your $OSTYPE package manager."
+			echo "Install GNU sed with your package manager."
 		fi
 		exit 1
 	fi
 }
 
+CLANG_FORMAT=""
+assign_clang_format() {
+	local -r binaries=(
+		"clang-format-18"
+		"clang-format-17"
+		"clang-format-16"
+		"clang-format-15"
+		"clang-format"
+	)
+
+	for binary in "${binaries[@]}"
+	do
+		if command -v "$binary" &> /dev/null; then
+			CLANG_FORMAT=$binary
+			break
+		fi
+	done
+
+	if [[ $CLANG_FORMAT == "" ]]; then
+		echo "Failed to find clang-format."
+		exit 1
+	fi
+}
+
 assert_min_version () {
 	if ! check_min_version "$1" "$2" ; then
 		echo "$3"
@@ -103,7 +130,7 @@ format () {
 	local -r since_ref=${1:-HEAD~1}
 	pushd "$(git rev-parse --show-toplevel)" > /dev/null
 	echo "Using paths relative to: $(pwd)"
-	find_cpp_files "$since_ref" | run_clang_format
+	find_cpp_files "$since_ref" | run_clang_format "$since_ref"
 	popd > /dev/null
 }
 
@@ -154,22 +181,24 @@ list_changed_files () {
 }
 
 run_clang_format () {
+	local -r since_ref=$1
 	while read -r src_file ; do
 		local ranges=()
 		while IFS=$'\n' read -r range ; do
 			ranges+=("$range")
-		done < <(git_diff_to_clang_line_range "$src_file")
+		done < <(git_diff_to_clang_line_range "$src_file" "$since_ref")
 
 		if (( "${#ranges[@]}" )); then
-			echo "clang-format -i ${ranges[*]} \"$src_file\""
-			clang-format -i "${ranges[@]}" "$src_file"
+			echo "$CLANG_FORMAT -i ${ranges[*]} \"$src_file\""
+			$CLANG_FORMAT -i "${ranges[@]}" "$src_file"
 		fi
 	done
 }
 
 git_diff_to_clang_line_range () {
 	local -r file=$1
-	git_diff --ignore-space-at-eol -U0 HEAD~1 "$file" \
+	local -r since_ref=$2
+	git_diff --ignore-space-at-eol -U0 "$since_ref" "$file" \
 		| grep -E "^@@" \
 		| filter_line_range \
 		| to_clang_line_range
