@@ -17,8 +17,12 @@ set -euo pipefail
 list_markdown_files () {
 	git ls-files -- \
 	  '*.md' \
+	  ':!:.github/*.md' \
+	  ':!:.github/**/*.md' \
 	  ':!:src/libs/*.md' \
-	  ':!:contrib/resources/*.md'
+	  ':!:src/hardware/reelmagic/docs/*.md' \
+	  ':!:contrib/resources/*.md' \
+	  ':!:contrib/resources/translations/*.md'
 }
 
 main () {
