@@ -1,4 +1,5 @@
 /*
+ *  Copyright (C) 2024-2024  The DOSBox Staging Team
  *  Copyright (C) 2002-2021  The DOSBox Team
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -17,16 +18,22 @@
  */
 #include "dosbox.h"
 
+// Needed for std::isnan in simde
+#include <cmath>
+
 #include "callback.h"
 #include "cpu.h"
 #include "fpu.h"
 #include "inout.h"
 #include "lazyflags.h"
 #include "mem.h"
+#include "mmx.h"
 #include "paging.h"
 #include "pic.h"
 #include "tracy.h"
 
+#include "simde/x86/mmx.h"
+
 #if C_DEBUG
 #include "debug.h"
 #endif
@@ -35,10 +42,12 @@
 #define LoadMb(off) mem_readb(off)
 #define LoadMw(off) mem_readw(off)
 #define LoadMd(off) mem_readd(off)
+#define LoadMq(off) mem_readq(off)
 
 #define SaveMb(off,val)	mem_writeb(off,val)
 #define SaveMw(off,val)	mem_writew(off,val)
 #define SaveMd(off,val)	mem_writed(off,val)
+#define SaveMq(off,val) mem_writeq(off,val)
 
 extern Bitu cycle_count;
 
@@ -133,7 +142,8 @@ static inline uint32_t Fetchd() {
 
 #define EALookupTable (core.ea_table)
 
-Bits CPU_Core_Simple_Run(void) {
+Bits CPU_Core_Simple_Run() noexcept
+{
 	ZoneScoped;
 	while (CPU_Cycles-->0) {
 		LOADIP;
@@ -195,7 +205,8 @@ decode_end:
 	return CBRET_NONE;
 }
 
-Bits CPU_Core_Simple_Trap_Run(void) {
+Bits CPU_Core_Simple_Trap_Run() noexcept
+{
 	Bits oldCycles = CPU_Cycles;
 	CPU_Cycles = 1;
 	cpu.trap_skip = false;
@@ -208,8 +219,6 @@ Bits CPU_Core_Simple_Trap_Run(void) {
 	return ret;
 }
 
-
-
 void CPU_Core_Simple_Init(void) {
 
 }
