@@ -1,7 +1,7 @@
 /*
  *  SPDX-License-Identifier: GPL-2.0-or-later
  *
- *  Copyright (C) 2020-2022  The DOSBox Staging Team
+ *  Copyright (C) 2020-2024  The DOSBox Staging Team
  *  Copyright (C) 2002-2021  The DOSBox Team
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -41,22 +41,24 @@
 #include <cstring>
 #endif
 
+#include "channel_names.h"
 #include "drives.h"
 #include "fs_utils.h"
+#include "math_utils.h"
 #include "setup.h"
 #include "string_utils.h"
-#include "math_utils.h"
-
-using namespace std;
 
 // String maximums, local to this file
 #define MAX_LINE_LENGTH 512
 #define MAX_FILENAME_LENGTH 256
 
 // STL type shorteners, local to this file
-using track_iter       = vector<CDROM_Interface_Image::Track>::iterator;
-using track_const_iter = vector<CDROM_Interface_Image::Track>::const_iterator;
-using tracks_size_t    = vector<CDROM_Interface_Image::Track>::size_type;
+using track_iter       = std::vector<CDROM_Interface_Image::Track>::iterator;
+using track_const_iter = std::vector<CDROM_Interface_Image::Track>::const_iterator;
+using tracks_size_t    = std::vector<CDROM_Interface_Image::Track>::size_type;
+
+// Ensure the maximum allowed redbook bytes stays within the API type sizes
+static_assert(MAX_REDBOOK_BYTES <= UINT32_MAX);
 
 // Report bad seeks that would go beyond the end of the track
 bool CDROM_Interface_Image::TrackFile::offsetInsideTrack(const uint32_t offset)
@@ -99,7 +101,7 @@ CDROM_Interface_Image::BinaryFile::BinaryFile(const char *filename, bool &error)
         : TrackFile(BYTES_PER_RAW_REDBOOK_FRAME),
           file(nullptr)
 {
-	file = new ifstream(filename, ios::in | ios::binary);
+	file = new std::ifstream(filename, std::ios::in | std::ios::binary);
 	// If new fails, an exception is generated and scope leaves this constructor
 	error = file->fail();
 }
@@ -139,7 +141,7 @@ int CDROM_Interface_Image::BinaryFile::getLength()
 {
 	// Return our cached result if we've already been asked before
 	if (length_redbook_bytes < 0 && file) {
-		file->seekg(0, ios::end);
+		file->seekg(0, std::ios::end);
 		/**
 		 *  All read(..) operations involve an absolute position and
 		 *  this function isn't called in other threads, therefore
@@ -176,13 +178,13 @@ bool CDROM_Interface_Image::BinaryFile::seek(const uint32_t offset)
 	if (static_cast<uint32_t>(file->tellg()) == offset)
 		return true;
 
-	file->seekg(offset, ios::beg);
+	file->seekg(offset, std::ios::beg);
 
 	// If the first seek attempt failed, then try harder
 	if (file->fail()) {
-		file->clear();                 // clear fail and eof bits
-		file->seekg(0, std::ios::beg); // "I have returned."
-		file->seekg(offset, ios::beg); // "It will be done."
+		file->clear();                      // clear fail and eof bits
+		file->seekg(0, std::ios::beg);      // "I have returned."
+		file->seekg(offset, std::ios::beg); // "It will be done."
 	}
 	return !file->fail();
 }
@@ -377,7 +379,6 @@ bool CDROM_Interface_Image::AudioFile::read(uint8_t *buffer,
 	// If the track is mono, convert to stereo
 	if (channels == 1 && decoded_frames) {
 #ifdef DEBUG
-		using namespace std::chrono;
 		using clock = std::chrono::steady_clock;
 		clock::time_point begin = clock::now(); // start the timer
 #endif
@@ -478,30 +479,30 @@ int CDROM_Interface_Image::AudioFile::getLength()
 
 // initialize static members
 int CDROM_Interface_Image::refCount = 0;
-CDROM_Interface_Image* CDROM_Interface_Image::images[26] = {};
 CDROM_Interface_Image::imagePlayer CDROM_Interface_Image::player;
 
-CDROM_Interface_Image::CDROM_Interface_Image(uint8_t sub_unit)
+CDROM_Interface_Image::CDROM_Interface_Image()
         : tracks{},
           readBuffer{},
           mcn("")
 {
-	images[sub_unit] = this;
 	if (refCount == 0) {
 		if (!player.channel) {
+			MIXER_LockMixerThread();
 			const auto mixer_callback = std::bind(&CDROM_Interface_Image::CDAudioCallBack,
 			                                      this, std::placeholders::_1);
 
 			player.channel = MIXER_AddChannel(mixer_callback,
-			                                  use_mixer_rate,
-			                                  "CDAUDIO",
+			                                  UseMixerRate,
+			                                  ChannelName::CdAudio,
 			                                  {ChannelFeature::Stereo,
 			                                   ChannelFeature::DigitalAudio});
 
 			player.channel->Enable(false); // only enabled during playback periods
+			MIXER_UnlockMixerThread();
 		}
 #ifdef DEBUG
-		LOG_MSG("CDROM: Initialized the CDAUDIO audio channel");
+		LOG_MSG("CDROM: Initialised the %s audio channel", ChannelName::CdAudio);
 #endif
 	}
 	refCount++;
@@ -509,24 +510,28 @@ CDROM_Interface_Image::CDROM_Interface_Image(uint8_t sub_unit)
 
 CDROM_Interface_Image::~CDROM_Interface_Image()
 {
+	MIXER_LockMixerThread();
 	refCount--;
 
 	// Stop playback before wiping out the CD Player
-	if (refCount == 0 && player.cd) {
-		StopAudio();
+	if (refCount == 0) {
+		LOG_MSG("CDROM: Shutting down CD-DA player");
+
+		if (player.cd) {
+			StopAudio();
+		}
+		MIXER_DeregisterChannel(player.channel);
 		player.channel.reset();
-#ifdef DEBUG
-		LOG_MSG("CDROM: Released CD Player resources");
-#endif
 	}
 	if (player.cd == this) {
 		player.cd = nullptr;
 	}
+	MIXER_UnlockMixerThread();
 }
 
-bool CDROM_Interface_Image::SetDevice(const char* path, [[maybe_unused]] const int cd_number)
+bool CDROM_Interface_Image::SetDevice(const char* path)
 {
-	const bool result = LoadCueSheet((char *)path) || LoadIsoFile((char *)path);
+	const bool result = LoadCueSheet(path) || LoadIsoFile(path);
 	if (!result) {
 		// print error message on dosbox console
 		char buf[MAX_LINE_LENGTH];
@@ -537,10 +542,10 @@ bool CDROM_Interface_Image::SetDevice(const char* path, [[maybe_unused]] const i
 	return result;
 }
 
-bool CDROM_Interface_Image::GetUPC(unsigned char& attr, char* upc)
+bool CDROM_Interface_Image::GetUPC(unsigned char& attr, std::string& upc)
 {
 	attr = 0;
-	strcpy(upc, this->mcn.c_str());
+	upc = mcn;
 #ifdef DEBUG
 	LOG_MSG("CDROM: GetUPC => returned %s", upc);
 #endif
@@ -630,6 +635,7 @@ bool CDROM_Interface_Image::GetAudioSub(unsigned char& attr,
 		 // reserve the track_file as a shared_ptr to avoid deletion in another thread
 		const auto track_file = player.trackFile.lock();
 		if (track_file) {
+			LagDriveResponse();
 			const uint32_t sample_rate = track_file->getRate();
 			const uint32_t played_frames = ceil_udivide(player.playedTrackFrames
 			                               * REDBOOK_FRAMES_PER_SECOND, sample_rate);
@@ -751,11 +757,13 @@ bool CDROM_Interface_Image::PlayAudioSector(uint32_t start, uint32_t len)
 
 	// Assign the mixer function associated with this track's content type
 	if (track_file->getEndian() == AUDIO_S16SYS) {
-		player.addFrames = track_channels ==  2  ? &MixerChannel::AddSamples_s16 \
-		                                         : &MixerChannel::AddSamples_m16;
+		player.addFrames = (track_channels == 2)
+		                         ? &MixerChannel::AddSamples_s16
+		                         : &MixerChannel::AddSamples_m16;
 	} else {
-		player.addFrames = track_channels ==  2  ? &MixerChannel::AddSamples_s16_nonnative \
-		                                         : &MixerChannel::AddSamples_m16_nonnative;
+		player.addFrames = (track_channels == 2)
+		                         ? &MixerChannel::AddSamples_s16_nonnative
+		                         : &MixerChannel::AddSamples_m16_nonnative;
 	}
 
 	/**
@@ -802,8 +810,9 @@ bool CDROM_Interface_Image::PlayAudioSector(uint32_t start, uint32_t len)
 bool CDROM_Interface_Image::PauseAudio(bool resume)
 {
 	player.isPaused = !resume;
-	if (player.channel)
+	if (player.channel) {
 		player.channel->Enable(resume);
+	}
 #ifdef DEBUG
 	LOG_MSG("CDROM: PauseAudio => audio is now %s",
 	        resume ? "unpaused" : "paused");
@@ -815,8 +824,9 @@ bool CDROM_Interface_Image::StopAudio(void)
 {
 	player.isPlaying = false;
 	player.isPaused = false;
-	if (player.channel)
+	if (player.channel) {
 		player.channel->Enable(false);
+	}
 #ifdef DEBUG
 	LOG_MSG("CDROM: StopAudio => stopped playback and halted the mixer");
 #endif
@@ -833,15 +843,13 @@ void CDROM_Interface_Image::ChannelControl(TCtrl ctrl)
 #endif
 		return;
 	}
-
 	// Adjust the volume of our mixer channel as defined by the application
-	player.channel->SetVolumeScale(static_cast<float>(ctrl.vol[0] / 255.0), // left vol
-	                               static_cast<float>(ctrl.vol[1] / 255.0)); // right vol
+	player.channel->SetAppVolume({ctrl.vol[0] / 255.0f, ctrl.vol[1] / 255.0f});
 
 	// Map the audio channels in our mixer channel as defined by the application
-	const auto left_mapped = static_cast<LINE_INDEX>(ctrl.out[0]);
-	const auto right_mapped = static_cast<LINE_INDEX>(ctrl.out[1]);
-	player.channel->ChangeChannelMap(left_mapped, right_mapped);
+	const auto left_mapped = static_cast<LineIndex>(ctrl.out[0]);
+	const auto right_mapped = static_cast<LineIndex>(ctrl.out[1]);
+	player.channel->SetChannelMap({left_mapped, right_mapped});
 
 #ifdef DEBUG
 	LOG_MSG("CDROM: ChannelControl => volumes %d/255 and %d/255, "
@@ -993,7 +1001,7 @@ bool CDROM_Interface_Image::ReadSectorsHost(void *buffer, bool raw, unsigned lon
 	return success;
 }
 
-void CDROM_Interface_Image::CDAudioCallBack(uint16_t desired_track_frames)
+void CDROM_Interface_Image::CDAudioCallBack(const int desired_track_frames)
 {
 	/**
 	 *  This callback runs in SDL's mixer thread, so there's a risk
@@ -1058,14 +1066,14 @@ void CDROM_Interface_Image::CDAudioCallBack(uint16_t desired_track_frames)
 	}
 }
 
-bool CDROM_Interface_Image::LoadIsoFile(char* filename)
+bool CDROM_Interface_Image::LoadIsoFile(const char* filename)
 {
 	tracks.clear();
 
 	// data track (track 1)
-	Track track;
-	bool error;
-	track.file = make_shared<BinaryFile>(filename, error);
+	Track track = {};
+	bool error  = false;
+	track.file  = std::make_shared<BinaryFile>(filename, error);
 
 	if (error) {
 		return false;
@@ -1133,7 +1141,8 @@ bool CDROM_Interface_Image::CanReadPVD(TrackFile *file,
 }
 
 #if defined(WIN32)
-static string dirname(char * file) {
+static std::string dirname(char* file)
+{
 	char * sep = strrchr(file, '\\');
 	if (sep == nullptr)
 		sep = strrchr(file, '/');
@@ -1148,7 +1157,7 @@ static string dirname(char * file) {
 }
 #endif
 
-bool CDROM_Interface_Image::LoadCueSheet(char *cuefile)
+bool CDROM_Interface_Image::LoadCueSheet(const char *cuefile)
 {
 	tracks.clear();
 
@@ -1162,9 +1171,9 @@ bool CDROM_Interface_Image::LoadCueSheet(char *cuefile)
 	bool canAddTrack = false;
 	char tmp[MAX_FILENAME_LENGTH];  // dirname can change its argument
 	safe_strcpy(tmp, cuefile);
-	string pathname(dirname(tmp));
-	ifstream in;
-	in.open(to_native_path(cuefile), ios::in);
+	std::string pathname(dirname(tmp));
+	std::ifstream in;
+	in.open(to_native_path(cuefile), std::ios::in);
 	if (in.fail()) {
 		return false;
 	}
@@ -1176,9 +1185,9 @@ bool CDROM_Interface_Image::LoadCueSheet(char *cuefile)
 		if (in.fail() && !in.eof()) {
 			return false;  // probably a binary file
 		}
-		istringstream line(buf);
+		std::istringstream line(buf);
 
-		string command;
+		std::string command;
 		GetCueKeyword(command, line);
 
 		if (command == "TRACK") {
@@ -1192,7 +1201,7 @@ bool CDROM_Interface_Image::LoadCueSheet(char *cuefile)
 
 			line >> track_number; // (cin) read into a true int first
 			track.number = static_cast<uint8_t>(track_number);
-			string type;
+			std::string type;
 			GetCueKeyword(type, line);
 
 			if (type == "AUDIO") {
@@ -1234,18 +1243,19 @@ bool CDROM_Interface_Image::LoadCueSheet(char *cuefile)
 			else success = true;
 			canAddTrack = false;
 
-			string filename;
+			std::string filename;
 			GetCueString(filename, line);
 			GetRealFileName(filename, pathname);
-			string type;
+			std::string type;
 			GetCueKeyword(type, line);
 
 			bool error = true;
 			if (type == "BINARY") {
-				track.file = make_shared<BinaryFile>(filename.c_str(), error);
-			}
-			else {
-				track.file = make_shared<AudioFile>(filename.c_str(), error);
+				track.file = std::make_shared<BinaryFile>(
+				        filename.c_str(), error);
+			} else {
+				track.file = std::make_shared<AudioFile>(
+				        filename.c_str(), error);
 				/**
 				 *  SDL_Sound first tries using a decoder having a matching
 				 *  registered extension as the filename, and then falls back to
@@ -1362,7 +1372,7 @@ bool CDROM_Interface_Image::AddTrack(Track &curr,
 	return true;
 }
 
-bool CDROM_Interface_Image::HasDataTrack(void)
+bool CDROM_Interface_Image::HasDataTrack() const
 {
 	//Data track has attribute 0x40
 	for (const auto &track : tracks) {
@@ -1374,7 +1384,7 @@ bool CDROM_Interface_Image::HasDataTrack(void)
 }
 
 
-bool CDROM_Interface_Image::GetRealFileName(string &filename, string &pathname)
+bool CDROM_Interface_Image::GetRealFileName(std::string &filename, std::string &pathname)
 {
 	// check if file exists
 	if (path_exists(filename)) {
@@ -1402,11 +1412,11 @@ bool CDROM_Interface_Image::GetRealFileName(string &filename, string &pathname)
 		return false;
 	}
 
-	localDrive *ldp = dynamic_cast<localDrive*>(Drives[drive]);
+	const auto ldp = std::dynamic_pointer_cast<localDrive>(Drives.at(drive));
 	if (ldp) {
-		ldp->GetSystemFilename(tmp, fullname);
-		if (path_exists(tmp)) {
-			filename = tmp;
+		const std::string host_filename = ldp->MapDosToHostFilename(fullname);
+		if (path_exists(host_filename)) {
+			filename = host_filename;
 			return true;
 		}
 	}
@@ -1414,7 +1424,7 @@ bool CDROM_Interface_Image::GetRealFileName(string &filename, string &pathname)
 	return false;
 }
 
-bool CDROM_Interface_Image::GetCueKeyword(string &keyword, istream &in)
+bool CDROM_Interface_Image::GetCueKeyword(std::string& keyword, std::istream& in)
 {
 	in >> keyword;
 	for (Bitu i = 0; i < keyword.size(); i++) {
@@ -1423,7 +1433,7 @@ bool CDROM_Interface_Image::GetCueKeyword(string &keyword, istream &in)
 	return true;
 }
 
-bool CDROM_Interface_Image::GetCueFrame(uint32_t &frames, istream &in)
+bool CDROM_Interface_Image::GetCueFrame(uint32_t& frames, std::istream& in)
 {
 	std::string msf;
 	in >> msf;
@@ -1433,15 +1443,15 @@ bool CDROM_Interface_Image::GetCueFrame(uint32_t &frames, istream &in)
 	return success;
 }
 
-bool CDROM_Interface_Image::GetCueString(string &str, istream &in)
+bool CDROM_Interface_Image::GetCueString(std::string& str, std::istream& in)
 {
 	int pos = (int)in.tellg();
 	in >> str;
 	if (str[0] == '\"') {
 		if (str[str.size() - 1] == '\"') {
-			str.assign(str, 1, str.size() - 2);
+			str = str.substr(1, str.size() - 2);
 		} else {
-			in.seekg(pos, ios::beg);
+			in.seekg(pos, std::ios::beg);
 			char buffer[MAX_FILENAME_LENGTH];
 			in.getline(buffer, MAX_FILENAME_LENGTH, '\"');	// skip
 			in.getline(buffer, MAX_FILENAME_LENGTH, '\"');
@@ -1455,9 +1465,10 @@ void CDROM_Image_Destroy(Section*) {
 	Sound_Quit();
 }
 
-void CDROM_Image_Init(Section* sec) {
+void CDROM_Image_Init(Section* sec)
+{
 	if (sec != nullptr) {
-		sec->AddDestroyFunction(CDROM_Image_Destroy, false);
+		sec->AddDestroyFunction(CDROM_Image_Destroy);
 	}
 	Sound_Init();
 }
