@@ -39,6 +39,7 @@
 #include "dos/dos_locale.h"
 #include "dos_inc.h"
 #include "hardware.h"
+#include "hardware/voodoo.h"
 #include "inout.h"
 #include "ints/int10.h"
 #include "mapper.h"
@@ -85,7 +86,6 @@ void FPU_Init(Section*);
 void DMA_Init(Section*);
 
 void PCI_Init(Section*);
-void VOODOO_Init(Section*);
 void VIRTUALBOX_Init(Section*);
 void VMWARE_Init(Section*);
 
@@ -297,7 +297,7 @@ static void increase_ticks()
 		auto ratio = (ticks.scheduled * (CPU_CyclePercUsed * 1024 / 100)) /
 		             ticks.done;
 
-		auto new_cycle_max = CPU_CycleMax.load();
+		auto new_cycle_max = CPU_CycleMax;
 
 		auto cproc = static_cast<int64_t>(CPU_CycleMax) *
 		             static_cast<int64_t>(ticks.scheduled);
@@ -417,7 +417,7 @@ static void increase_ticks()
 		// ticks.added > 15 but ticks.scheduled < 5, lower the cycles
 		// but do not reset the scheduled/done ticks to take them into
 		// account during the next auto cycle adjustment.
-		CPU_CycleMax = CPU_CycleMax / 3;
+		CPU_CycleMax /= 3;
 
 		if (CPU_CycleMax < auto_cpu_cycles_min) {
 			CPU_CycleMax = auto_cpu_cycles_min;
@@ -462,7 +462,7 @@ static void DOSBOX_UnlockSpeed( bool pressed ) {
 		if (CPU_CycleAutoAdjust) {
 			autoadjust = true;
 			CPU_CycleAutoAdjust = false;
-			CPU_CycleMax = CPU_CycleMax / 3;
+			CPU_CycleMax /= 3;
 			if (CPU_CycleMax<1000) CPU_CycleMax=1000;
 		}
 	} else {
@@ -652,7 +652,7 @@ void DOSBOX_Init()
 	secprop->AddInitFunction(&PAGING_Init);
 	secprop->AddInitFunction(&MEM_Init);
 
-	pint = secprop->Add_int("memsize", when_idle, 16);
+	pint = secprop->Add_int("memsize", only_at_start, 16);
 	pint->SetMinMax(MEM_GetMinMegabytes(), MEM_GetMaxMegabytes());
 	pint->Set_help(
 	        "Amount of memory of the emulated machine has in MB (16 by default).\n"
@@ -820,37 +820,8 @@ void DOSBOX_Init()
 	secprop->AddInitFunction(&KEYBOARD_Init);
 	secprop->AddInitFunction(&PCI_Init); // PCI bus
 
-	secprop = control->AddSection_prop("voodoo", &VOODOO_Init);
-
-	pbool = secprop->Add_bool("voodoo", when_idle, true);
-	pbool->Set_help("Enable 3dfx Voodoo emulation (enabled by default).");
-
-	pstring = secprop->Add_string("voodoo_memsize", only_at_start, "4");
-	pstring->Set_values({"4", "12"});
-	pstring->Set_help(
-	        "Set the amount of video memory for 3dfx Voodoo graphics, either 4 or 12 MB.\n"
-	        "The memory is used by the Frame Buffer Interface (FBI) and Texture Mapping Unit\n"
-	        "(TMU) as follows:\n"
-	        "   4: 2 MB for the FBI and one TMU with 2 MB (default).\n"
-	        "  12: 4 MB for the FBI and two TMUs, each with 4 MB.");
-
-	// Deprecate the boolean Voodoo multithreading setting
-	pbool = secprop->Add_bool("voodoo_multithreading", deprecated, false);
-	pbool->Set_help("Renamed to 'voodoo_threads'");
-
-	pstring = secprop->Add_string("voodoo_threads", only_at_start, "auto");
-	pstring->Set_help(
-	        "Use threads to improve 3dfx Voodoo performance:\n"
-	        "  auto:     Use up to 8 threads based on available CPU cores (default).\n"
-	        "  <value>:  Set a specific number of threads between 1 and 16.\n"
-	        "Note: Tests show that frame rates increase up to 8 threads after which\n"
-	        "      they level off or decrease, in general.");
-
-	pbool = secprop->Add_bool("voodoo_bilinear_filtering", only_at_start, false);
-	pbool->Set_help(
-	        "Use bilinear filtering to emulate the 3dfx Voodoo's texture smoothing effect\n"
-	        "(disabled by default). Only suggested if you have a fast desktop-class CPU, as\n"
-	        "it can impact frame rates on slower systems.");
+	// Configure 3dfx Voodoo settings
+	VOODOO_AddConfigSection(control);
 
 	// Configure capture
 	CAPTURE_AddConfigSection(control);
@@ -1119,7 +1090,7 @@ void DOSBOX_Init()
 	        "Can be disabled, dummy, mouse, modem, nullmodem, direct ('dummy' by default).\n"
 	        "Additional parameters must be on the same line in the form of\n"
 	        "parameter:value. The optional 'irq' parameter is common for all types.\n"
-	        "  - for 'mouse':      model (overrides the setting from the [mouse] section)\n"
+	        "  - for 'mouse':      model (optional; overrides the 'com_mouse_model' setting).\n"
 	        "  - for 'direct':     realport (required), rxdelay (optional).\n"
 	        "                      (e.g., realport:COM1, realport:ttyS0).\n"
 	        "  - for 'modem':      listenport, sock, bps (all optional).\n"
@@ -1239,6 +1210,15 @@ void DOSBOX_Init()
 	        "tab-separated format, used by SETVER.EXE as a persistent storage\n"
 	        "(empty by default).");
 
+	secprop->AddInitFunction(&DOS_InitFileLocking, changeable_at_runtime);
+	pbool = secprop->Add_bool("file_locking", when_idle, true);
+	pbool->Set_help(
+	        "Enable file locking (SHARE.EXE emulation; enabled by default).\n"
+	        "This is required for some Windows 3.1x applications to work properly.\n"
+	        "It generally does not cause problems for DOS games except in rare cases\n"
+	        "(e.g., Astral Blur demo). If you experience crashes related to file\n"
+	        "permissions, you can try disabling this.");
+
 	// Mscdex
 	secprop->AddInitFunction(&MSCDEX_Init);
 	secprop->AddInitFunction(&DRIVES_Init);
@@ -1354,7 +1334,7 @@ void DOSBOX_Init()
 	        "Important: The [autoexec] section must be the last section in the config!");
 
 	MSG_Add("CONFIGFILE_INTRO",
-	        "# This is the configuration file for " DOSBOX_PROJECT_NAME
+	        "# This is the configuration file for " DOSBOX_NAME
 	        " (%s).\n"
 	        "# Lines starting with a '#' character are comments.\n");
 
