@@ -1,7 +1,7 @@
 /*
  *  SPDX-License-Identifier: GPL-2.0-or-later
  *
- *  Copyright (C) 2022-2022  The DOSBox Staging Team
+ *  Copyright (C) 2022-2024  The DOSBox Staging Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -25,15 +25,17 @@
 
 CHECK_NARROWING();
 
+// #define DEBUG_ADLIB_GOLD
+
 // Yamaha YM7128B Surround Processor emulation
 // -------------------------------------------
 
-SurroundProcessor::SurroundProcessor(const uint16_t sample_rate)
+SurroundProcessor::SurroundProcessor(const int sample_rate_hz)
 {
-	assert(sample_rate >= 10);
+	assert(sample_rate_hz >= 10);
 
 	YM7128B_ChipIdeal_Ctor(&chip);
-	YM7128B_ChipIdeal_Setup(&chip, sample_rate);
+	YM7128B_ChipIdeal_Setup(&chip, sample_rate_hz);
 	YM7128B_ChipIdeal_Reset(&chip);
 	YM7128B_ChipIdeal_Start(&chip);
 }
@@ -57,9 +59,11 @@ void SurroundProcessor::ControlWrite(const uint8_t val)
 
 	// Change register data at the falling edge of 'a0' word clock
 	if (control_state.a0 && !reg.a0) {
-		//		DEBUG_LOG_MSG("ADLIBGOLD: Surround: Write
-		// control register %d, data: %d",
-		// control_state.addr, control_state.data);
+#ifdef DEBUG_ADLIB_GOLD
+		LOG_DEBUG("ADLIBGOLD: Surround: Write control register %d, data: %d",
+		          control_state.addr,
+		          control_state.data);
+#endif
 
 		YM7128B_ChipIdeal_Write(&chip, control_state.addr, control_state.data);
 	} else {
@@ -68,16 +72,17 @@ void SurroundProcessor::ControlWrite(const uint8_t val)
 		// the rising edge of 'sci'.
 		if (!control_state.sci && reg.sci) {
 			// The 'a0' word clock determines the type of the data.
-			if (reg.a0)
+			if (reg.a0) {
 				// Data cycle
 				control_state.data = static_cast<uint8_t>(
 				                             control_state.data << 1) |
 				                     reg.din;
-			else
+			} else {
 				// Address cycle
 				control_state.addr = static_cast<uint8_t>(
 				                             control_state.addr << 1) |
 				                     reg.din;
+			}
 		}
 	}
 
@@ -85,7 +90,7 @@ void SurroundProcessor::ControlWrite(const uint8_t val)
 	control_state.a0  = reg.a0;
 }
 
-AudioFrame SurroundProcessor::Process(const AudioFrame &frame)
+AudioFrame SurroundProcessor::Process(const AudioFrame frame)
 {
 	YM7128B_ChipIdeal_Process_Data data = {};
 
@@ -99,32 +104,34 @@ AudioFrame SurroundProcessor::Process(const AudioFrame &frame)
 // Philips Semiconductors TDA8425 hi-fi stereo audio processor emulation
 // ---------------------------------------------------------------------
 
-StereoProcessor::StereoProcessor(const uint16_t _sample_rate)
-        : sample_rate(_sample_rate)
+StereoProcessor::StereoProcessor(const int _sample_rate_hz)
+        : sample_rate_hz(_sample_rate_hz)
 {
-	assert(sample_rate > 0);
+	assert(sample_rate_hz > 0);
 
-	constexpr auto allpass_freq = 400.0;
-	constexpr auto q_factor     = 1.7;
-	allpass.setup(sample_rate, allpass_freq, q_factor);
+	constexpr auto allpass_freq_hz = 400.0;
+	constexpr auto q_factor        = 1.7;
+	allpass.setup(sample_rate_hz, allpass_freq_hz, q_factor);
 
 	Reset();
 }
 
 void StereoProcessor::SetLowShelfGain(const double gain_db)
 {
-	constexpr auto cutoff_freq = 400.0;
-	constexpr auto slope       = 0.5;
-	for (auto &f : lowshelf)
-		f.setup(sample_rate, cutoff_freq, gain_db, slope);
+	constexpr auto cutoff_freq_hz = 400.0;
+	constexpr auto slope          = 0.5;
+	for (auto& f : lowshelf) {
+		f.setup(sample_rate_hz, cutoff_freq_hz, gain_db, slope);
+	}
 }
 
 void StereoProcessor::SetHighShelfGain(const double gain_db)
 {
-	constexpr auto cutoff_freq = 2500.0;
-	constexpr auto slope       = 0.5;
-	for (auto &f : highshelf)
-		f.setup(sample_rate, cutoff_freq, gain_db, slope);
+	constexpr auto cutoff_freq_hz = 2500.0;
+	constexpr auto slope          = 0.5;
+	for (auto& f : highshelf) {
+		f.setup(sample_rate_hz, cutoff_freq_hz, gain_db, slope);
+	}
 }
 
 StereoProcessor::~StereoProcessor() = default;
@@ -140,8 +147,10 @@ void StereoProcessor::Reset()
 	ControlWrite(StereoProcessorControlReg::Treble, shelf_filter_0db_value);
 
 	StereoProcessorSwitchFunctions sf = {};
-	sf.source_selector                = static_cast<uint8_t>(
-                StereoProcessorSourceSelector::Stereo1);
+
+	sf.source_selector = static_cast<uint8_t>(
+	        StereoProcessorSourceSelector::Stereo1);
+
 	sf.stereo_mode = static_cast<uint8_t>(StereoProcessorStereoMode::LinearStereo);
 
 	ControlWrite(StereoProcessorControlReg::SwitchFunctions, sf.data);
@@ -179,17 +188,23 @@ void StereoProcessor::ControlWrite(const StereoProcessorControlReg reg,
 	case StereoProcessorControlReg::VolumeLeft: {
 		const auto value = data & volume_control_mask;
 		gain.left        = calc_volume_gain(value);
-		DEBUG_LOG_MSG("ADLIBGOLD: Stereo: Final left volume set to %.2fdB (value %d)",
-		              gain.left,
-		              value);
+
+#ifdef DEBUG_ADLIB_GOLD
+		LOG_DEBUG("ADLIBGOLD: Stereo: Final left volume set to %.2fdB (value %d)",
+		          gain.left,
+		          value);
+#endif
 	} break;
 
 	case StereoProcessorControlReg::VolumeRight: {
 		const auto value = data & volume_control_mask;
 		gain.right       = calc_volume_gain(value);
-		DEBUG_LOG_MSG("ADLIBGOLD: Stereo: Final right volume set to %.2fdB (value %d)",
-		              gain.right,
-		              value);
+
+#ifdef DEBUG_ADLIB_GOLD
+		LOG_DEBUG("ADLIBGOLD: Stereo: Final right volume set to %.2fdB (value %d)",
+		          gain.right,
+		          value);
+#endif
 	} break;
 
 	case StereoProcessorControlReg::Bass: {
@@ -197,9 +212,11 @@ void StereoProcessor::ControlWrite(const StereoProcessorControlReg reg,
 		const auto gain_db = calc_filter_gain_db(value);
 		SetLowShelfGain(gain_db);
 
-		DEBUG_LOG_MSG("ADLIBGOLD: Stereo: Bass gain set to %.2fdB (value %d)",
-		              gain_db,
-		              value);
+#ifdef DEBUG_ADLIB_GOLD
+		LOG_DEBUG("ADLIBGOLD: Stereo: Bass gain set to %.2fdB (value %d)",
+		          gain_db,
+		          value);
+#endif
 	} break;
 
 	case StereoProcessorControlReg::Treble: {
@@ -210,9 +227,11 @@ void StereoProcessor::ControlWrite(const StereoProcessorControlReg reg,
 		const auto gain_db = calc_filter_gain_db(value + extra_treble);
 		SetHighShelfGain(gain_db);
 
-		DEBUG_LOG_MSG("ADLIBGOLD: Stereo: Treble gain set to %.2fdB (value %d)",
-		              gain_db,
-		              value);
+#ifdef DEBUG_ADLIB_GOLD
+		LOG_DEBUG("ADLIBGOLD: Stereo: Treble gain set to %.2fdB (value %d)",
+		          gain_db,
+		          value);
+#endif
 	} break;
 
 	case StereoProcessorControlReg::SwitchFunctions: {
@@ -223,14 +242,16 @@ void StereoProcessor::ControlWrite(const StereoProcessorControlReg reg,
 		stereo_mode = StereoProcessorStereoMode(
 		        static_cast<uint8_t>(sf.stereo_mode));
 
-		DEBUG_LOG_MSG("ADLIBGOLD: Stereo: Source selector set to %d, stereo mode set to %d",
-		              static_cast<int>(source_selector),
-		              static_cast<int>(stereo_mode));
+#ifdef DEBUG_ADLIB_GOLD
+		LOG_DEBUG("ADLIBGOLD: Stereo: Source selector set to %d, stereo mode set to %d",
+		          static_cast<int>(source_selector),
+		          static_cast<int>(stereo_mode));
+#endif
 	} break;
 	}
 }
 
-AudioFrame StereoProcessor::ProcessSourceSelection(const AudioFrame &frame)
+AudioFrame StereoProcessor::ProcessSourceSelection(const AudioFrame frame)
 {
 	switch (source_selector) {
 	case StereoProcessorSourceSelector::SoundA1:
@@ -250,7 +271,7 @@ AudioFrame StereoProcessor::ProcessSourceSelection(const AudioFrame &frame)
 	}
 }
 
-AudioFrame StereoProcessor::ProcessShelvingFilters(const AudioFrame &frame)
+AudioFrame StereoProcessor::ProcessShelvingFilters(const AudioFrame frame)
 {
 	AudioFrame out_frame = {};
 
@@ -261,7 +282,7 @@ AudioFrame StereoProcessor::ProcessShelvingFilters(const AudioFrame &frame)
 	return out_frame;
 }
 
-AudioFrame StereoProcessor::ProcessStereoProcessing(const AudioFrame &frame)
+AudioFrame StereoProcessor::ProcessStereoProcessing(const AudioFrame frame)
 {
 	AudioFrame out_frame = {};
 
@@ -292,7 +313,7 @@ AudioFrame StereoProcessor::ProcessStereoProcessing(const AudioFrame &frame)
 	return out_frame;
 }
 
-AudioFrame StereoProcessor::Process(const AudioFrame &frame)
+AudioFrame StereoProcessor::Process(const AudioFrame frame)
 {
 	auto out_frame = ProcessSourceSelection(frame);
 	out_frame      = ProcessShelvingFilters(out_frame);
@@ -307,12 +328,12 @@ AudioFrame StereoProcessor::Process(const AudioFrame &frame)
 // AdLib Gold module
 // -----------------
 
-AdlibGold::AdlibGold(const uint16_t sample_rate)
+AdlibGold::AdlibGold(const int sample_rate_hz)
         : surround_processor(nullptr),
           stereo_processor(nullptr)
 {
-	surround_processor = std::make_unique<SurroundProcessor>(sample_rate);
-	stereo_processor   = std::make_unique<StereoProcessor>(sample_rate);
+	surround_processor = std::make_unique<SurroundProcessor>(sample_rate_hz);
+	stereo_processor   = std::make_unique<StereoProcessor>(sample_rate_hz);
 }
 
 AdlibGold::~AdlibGold() = default;
@@ -328,7 +349,7 @@ void AdlibGold::SurroundControlWrite(const uint8_t val)
 	surround_processor->ControlWrite(val);
 }
 
-void AdlibGold::Process(const int16_t *in, const uint32_t frames, float *out)
+void AdlibGold::Process(const int16_t* in, const int frames, float* out)
 {
 	auto frames_remaining = frames;
 
