@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2022-2022  The DOSBox Staging Team
+ *  Copyright (C) 2022-2023  The DOSBox Staging Team
  *  Copyright (C) 2002-2021  The DOSBox Team
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -22,7 +22,7 @@
 
 #include "serialport.h"
 
-#include "../mouse/mouse_config.h"
+#include "../input/mouse_config.h"
 
 class CSerialMouse : public CSerial {
 public:
@@ -31,7 +31,7 @@ public:
 
 	void NotifyMoved(const float x_rel, const float y_rel);
 	void NotifyButton(const uint8_t new_buttons,
-	                  const uint8_t idx); // changed button, staring from 0
+	                  const MouseButtonId id); // changed button
 	void NotifyWheel(const int16_t w_rel);
 
 	void BoostRate(const uint16_t rate_hz); // 0 = standard rate
