@@ -20,18 +20,19 @@
 
 #if C_MODEM
 
-#include <ctype.h>
-#include <stdlib.h>
-#include <string.h>
-#include <utility>
+#include <cctype>
+#include <cstdlib>
+#include <cstring>
 #include <fstream>
 #include <sstream>
+#include <utility>
 
 #include "string_utils.h"
 #include "serialport.h"
 #include "softmodem.h"
 #include "math_utils.h"
 #include "misc_util.h"
+#include "version.h"
 
 class PhonebookEntry {
 public:
@@ -67,12 +68,13 @@ static bool MODEM_IsPhoneValid(const std::string &input) {
 	return true;
 }
 
-bool MODEM_ReadPhonebook(const std::string &filename) {
-	std::ifstream loadfile(filename);
+bool MODEM_ReadPhonebook(const std_fs::path &path) {
+	std::ifstream loadfile(path);
 	if (!loadfile)
 		return false;
 
-	LOG_MSG("SERIAL: Phonebook loading from %s.", filename.c_str());
+	const auto path_string = path.string();
+	LOG_MSG("SERIAL: Phonebook loading from %s.", path_string.c_str());
 
 	std::string linein;
 	while (std::getline(loadfile, linein)) {
@@ -80,8 +82,8 @@ bool MODEM_ReadPhonebook(const std::string &filename) {
 		std::string phone, address;
 
 		if (!(iss >> phone >> address)) {
-			LOG_MSG("SERIAL: Phonebook skipped a bad line in %s.",
-			        filename.c_str());
+			LOG_MSG("SERIAL: Phonebook skipped a bad line in '%s'",
+			        path_string.c_str());
 			continue;
 		}
 
@@ -125,7 +127,7 @@ CSerialModem::CSerialModem(const uint8_t port_idx, CommandLine *cmd)
 	// enet: Setting to 1 enables enet on the port, otherwise TCP.
 	if (getUintFromString("sock:", bool_temp, cmd)) {
 		if (bool_temp == 1) {
-			socketType = SOCKET_TYPE_ENET;
+			socketType = SocketType::Enet;
 		}
 	}
 
@@ -152,19 +154,14 @@ CSerialModem::CSerialModem(const uint8_t port_idx, CommandLine *cmd)
 	}
 
 	// Get the connect speed to report
-	constexpr auto min_baudrate = 300u;
-	constexpr auto max_baudrate = 57600u;
-	if (getUintFromString("baudrate:", val, cmd)) {
-		val = clamp(val, min_baudrate, max_baudrate);
+	uint32_t requested_bps = 0;
+	if (getUintFromString("bps:", val, cmd)) {
+		requested_bps = val;
 	} else {
-		val = max_baudrate;
+		requested_bps = 57600;
 	}
 
-	assert(val >= min_baudrate && val <= max_baudrate);
-	safe_sprintf(connect_string, "CONNECT %d", val);
-
-	LOG_MSG("SERIAL: Port %" PRIu8 " will report baud rate %d",
-			GetPortNumber(), val);
+	SetModemSpeed(requested_bps);
 
 	InstallationSuccessful=true;
 }
@@ -227,7 +224,7 @@ void CSerialModem::handleUpperEvent(uint16_t type)
 void CSerialModem::SendLine(const char *line) {
 	rqueue->addb(reg[MREG_CR_CHAR]);
 	rqueue->addb(reg[MREG_LF_CHAR]);
-	rqueue->adds((uint8_t *)line, strlen(line));
+	rqueue->adds((const uint8_t *)line, strlen(line));
 	rqueue->addb(reg[MREG_CR_CHAR]);
 	rqueue->addb(reg[MREG_LF_CHAR]);
 }
@@ -249,28 +246,30 @@ void CSerialModem::SendNumber(uint32_t val)
 }
 
 void CSerialModem::SendRes(const ResTypes response) {
-	char const * string = nullptr;
+	const char* response_str = nullptr;
 	uint32_t code = -1;
 	switch (response) {
-		case ResOK:         code = 0; string = "OK"; break;
-		case ResCONNECT:    code = 1; string = connect_string; break;
-		case ResRING:       code = 2; string = "RING"; break;
-		case ResNOCARRIER:  code = 3; string = "NO CARRIER"; break;
-		case ResERROR:      code = 4; string = "ERROR"; break;
-		case ResNODIALTONE: code = 6; string = "NO DIALTONE"; break;
-		case ResBUSY:       code = 7; string = "BUSY"; break;
-		case ResNOANSWER:   code = 8; string = "NO ANSWER"; break;
-		case ResNONE:       return;
+	case ResOK:         code = 0; response_str = "OK"; break;
+	case ResCONNECT:    code = 1; response_str = connect_string; break;
+	case ResRING:       code = 2; response_str = "RING"; break;
+	case ResNOCARRIER:  code = 3; response_str = "NO CARRIER"; break;
+	case ResERROR:      code = 4; response_str = "ERROR"; break;
+	case ResNODIALTONE: code = 6; response_str = "NO DIALTONE"; break;
+	case ResBUSY:       code = 7; response_str = "BUSY"; break;
+	case ResNOANSWER:   code = 8; response_str = "NO ANSWER"; break;
+	case ResNONE: return;
 	}
 
 	if(doresponse != 1) {
 		if (doresponse == 2 && (response == ResRING ||
-			response == ResCONNECT || response == ResNOCARRIER))
+			response == ResCONNECT || response == ResNOCARRIER)) {
 			return;
-		if (numericresponse && code != static_cast<uint32_t>(-1))
+		}
+		if (numericresponse && code != static_cast<uint32_t>(-1)) {
 			SendNumber(code);
-		else if (string != nullptr)
-			SendLine(string);
+		} else if (response_str != nullptr) {
+			SendLine(response_str);
+		}
 
 		// if(CSerial::CanReceiveByte())	// very fast response
 		//	if(rqueue->inuse() && CSerial::getRTS())
@@ -280,14 +279,22 @@ void CSerialModem::SendRes(const ResTypes response) {
 		//	        GetPortNumber(), rbyte);
 		//	}
 
-		if (string != nullptr) {
+		if (response_str != nullptr) {
 			LOG_MSG("SERIAL: Port %" PRIu8 " modem response: %s.",
-			        GetPortNumber(), string);
+			        GetPortNumber(),
+			        response_str);
 		}
 	}
 }
 
 bool CSerialModem::Dial(const char * host) {
+	// Close the server socket before dialing
+	waitingclientsocket.reset();
+	if (serversocket) {
+		serversocket->Close();
+		serversocket.reset();
+	}
+
 	char buf[128] = "";
 	safe_strcpy(buf, host);
 
@@ -310,7 +317,7 @@ bool CSerialModem::Dial(const char * host) {
 	clientsocket.reset(NETClientSocket::NETClientFactory(socketType,
 	                                                     destination, port));
 	if (!clientsocket->isopen) {
-		clientsocket.reset(nullptr);
+		clientsocket.reset();
 		LOG_MSG("SERIAL: Port %" PRIu8 " failed to connect.", GetPortNumber());
 		SendRes(ResNOCARRIER);
 		EnterIdleState();
@@ -351,6 +358,25 @@ char CSerialModem::GetChar(char * & scan) const {
 	return ch;
 }
 
+void CSerialModem::SetModemSpeed(const uint32_t cfg_val) {
+	modem_bps_config = cfg_val;
+	LOG_MSG("SERIAL: Port %" PRIu8 " modem will report connection speed "
+	        "of up to %d bits per second",
+	        GetPortNumber(),
+	        modem_bps_config);
+	UpdateConnectString();
+}
+
+void CSerialModem::UpdateConnectString() {
+	const uint32_t connect_val = clamp(modem_bps_config,
+	                                   SerialMinBaudRate,
+	                                   GetPortBaudRate());
+
+	assert(connect_val >= SerialMinBaudRate &&
+	       connect_val <= SerialMaxBaudRate);
+	safe_sprintf(connect_string, "CONNECT %d", connect_val);
+}
+
 void CSerialModem::Reset(){
 	EnterIdleState();
 	cmdpos = 0;
@@ -359,7 +385,7 @@ void CSerialModem::Reset(){
 	plusinc = 0;
 	oldDTRstate = getDTR();
 	dtrmode = 2;
-	clientsocket.reset(nullptr);
+	clientsocket.reset();
 
 	memset(&reg,0,sizeof(reg));
 	reg[MREG_AUTOANSWER_COUNT] = 0;  // no autoanswer
@@ -382,18 +408,16 @@ void CSerialModem::EnterIdleState(){
 	ringing = false;
 	dtrofftimer = -1;
 	warmup_remain_ticks = 0;
-	clientsocket.reset(nullptr);
-	waitingclientsocket.reset(nullptr);
 
 	// get rid of everything
+	clientsocket.reset();
+	waitingclientsocket.reset();
 	if (serversocket) {
-		waitingclientsocket.reset(serversocket->Accept());
-		while (waitingclientsocket) {
-			waitingclientsocket.reset(serversocket->Accept());
-		}
+		serversocket->Close();
+		serversocket.reset();
 	}
+
 	if (listenport) {
-		serversocket.reset(nullptr);
 		serversocket.reset(NETServerSocket::NETServerFactory(socketType,
 		                                                     listenport));
 		if (!serversocket->isopen) {
@@ -401,13 +425,12 @@ void CSerialModem::EnterIdleState(){
 			        "%" PRIu16 ".",
 			        GetPortNumber(), listenport);
 
-			serversocket.reset(nullptr);
+			serversocket.reset();
 		} else
 			LOG_MSG("SERIAL: Port %" PRIu8 " modem listening on port "
 			        "%" PRIu16 " ...",
 			        GetPortNumber(), listenport);
 	}
-	waitingclientsocket.reset(nullptr);
 
 	commandmode = true;
 	CSerial::setCD(false);
@@ -419,7 +442,7 @@ void CSerialModem::EnterIdleState(){
 
 void CSerialModem::EnterConnectedState() {
 	// we don't accept further calls
-	serversocket.reset(nullptr);
+	serversocket.reset();
 	SendRes(ResCONNECT);
 	commandmode = false;
 	telClient = {}; // reset values
@@ -493,21 +516,32 @@ void CSerialModem::DoCommand()
 			if (is_next_token("SOCK", scanbuf)) {
 				scanbuf += 4;
 				const uint32_t requested_mode = ScanNumber(scanbuf);
-				if (requested_mode >= SOCKET_TYPE_COUNT) {
+				const auto requested_type = static_cast<SocketType>(
+				        requested_mode);
+				if (requested_type >= SocketType::Invalid) {
 					SendRes(ResERROR);
 					return;
 				}
-				if (socketType != (SocketTypesE)requested_mode) {
-					socketType = (SocketTypesE)requested_mode;
-					// This will break when there's more than two socket types.
-					LOG_MSG("SERIAL: Port %" PRIu8 " socket type %s",
+				if (socketType != requested_type) {
+					socketType = requested_type;
+					// This will break when there's more
+					// than two socket types.
+					LOG_MSG("SERIAL: Port %" PRIu8
+					        " socket type %s",
 					        GetPortNumber(),
-					        socketType ? "ENet" : "TCP");
+					        to_string(socketType));
 					// Reset port state.
 					EnterIdleState();
 				}
 				break;
 			}
+			// Set reported connection speed.
+			if (is_next_token("BPS", scanbuf)) {
+				scanbuf += 3;
+				const auto requested_bps = ScanNumber(scanbuf);
+				SetModemSpeed(requested_bps);
+				break;
+			}
 			// If the command wasn't recognized then stop parsing
 			SendRes(ResERROR);
 			return;
@@ -547,7 +581,7 @@ void CSerialModem::DoCommand()
 					// be a valid IP/name
 					// Transform by adding dots
 					size_t j = 0;
-					size_t foundlen = strlen(foundstr);
+					const size_t foundlen = strlen(foundstr);
 					for (size_t i = 0; i < foundlen; i++) {
 						buffer[j++] = foundstr[i];
 						// Add a dot after the third, sixth and ninth number
@@ -555,7 +589,7 @@ void CSerialModem::DoCommand()
 							buffer[j++] = '.';
 						// If the string is longer than 12 digits,
 						// interpret the rest as port
-						if (i == 11 && strlen(foundstr) > 12)
+						if (i == 11 && foundlen > 12)
 							buffer[j++] = ':';
 					}
 					buffer[j] = 0;
@@ -580,8 +614,12 @@ void CSerialModem::DoCommand()
 		}
 		case 'I': // Some strings about firmware
 			switch (ScanNumber(scanbuf)) {
-			case 3: SendLine("DOSBox Emulated Modem Firmware V1.00"); break;
-			case 4: SendLine("Modem compiled for DOSBox version " VERSION); break;
+			case 3:
+				SendLine("DOSBox Staging Emulated Modem Firmware V1.00");
+				break;
+			case 4:
+				SendLine("Modem compiled for DOSBox version " DOSBOX_VERSION);
+				break;
 			}
 			break;
 		case 'E': // Echo on/off
@@ -953,6 +991,7 @@ void CSerialModem::Timer2()
 			EnterIdleState();
 		}
 	}
+
 	// Handle incoming to the serial port
 	if (!commandmode && clientsocket && rqueue->left()) {
 		size_t usesize = rqueue->left() >= 16 ? 16 : rqueue->left();
@@ -970,8 +1009,30 @@ void CSerialModem::Timer2()
 		}
 	}
 
+	// Discard any incoming traffic while in command mode
+	if (commandmode && clientsocket) {
+		// Read buffer size doesn't really matter, we just need big
+		// enough buffer to discard data decently quickly
+		size_t usesize = 80;
+		if (!clientsocket->ReceiveArray(tmpbuf, usesize)) {
+			SendRes(ResNOCARRIER);
+			LOG_INFO("SERIAL: No carrier on receive");
+			EnterIdleState();
+		}
+	}
+
+	// Discard any incoming traffic from the waiting client
+	if (waitingclientsocket) {
+		// Read buffer size doesn't really matter, we just need big
+		// enough buffer to discard data decently quickly
+		size_t usesize = 80;
+		if (!waitingclientsocket->ReceiveArray(tmpbuf, usesize)) {
+			EnterIdleState();
+		}
+	}
+
 	// Tick down warmup timer
-	if (clientsocket && warmup_remain_ticks) {
+	if (connected && warmup_remain_ticks) {
 		// Drop all incoming and outgoing traffic for a short period after
 		// answering a call. This is to simulate real modem behavior where
 		// the first packet is usually bad (extra data in the buffer from
@@ -1082,6 +1143,7 @@ void CSerialModem::transmitByte(uint8_t val, bool first)
 void CSerialModem::updatePortConfig(uint16_t, uint8_t lcr)
 {
 	(void)lcr; // deliberately unused but needed for API compliance
+	UpdateConnectString();
 }
 
 void CSerialModem::updateMSR() {
