@@ -1,7 +1,7 @@
 /*
  * This file is part of libsidplayfp, a SID player engine.
  *
- * Copyright 2011-2022 Leandro Nini <drfiemost@users.sourceforge.net>
+ * Copyright 2011-2023 Leandro Nini <drfiemost@users.sourceforge.net>
  * Copyright 2007-2010 Antti Lankila
  *
  * This program is free software; you can redistribute it and/or modify
@@ -108,11 +108,6 @@ private:
 private:
     WaveformCalculator();
 
-    /**
-     * Build waveform table.
-     */
-    void buildWaveTable();
-
 public:
     /**
      * Get the singleton instance.
