@@ -1,7 +1,7 @@
 /*
  *  SPDX-License-Identifier: GPL-2.0-or-later
  *
- *  Copyright (C) 2021-2022  The DOSBox Staging Team
+ *  Copyright (C) 2021-2023  The DOSBox Staging Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -25,13 +25,13 @@
 
 // During testing we never want to log to stdout/stderr, as it could
 // negatively affect test harness.
-void GFX_ShowMsg(const char *, ...) {}
+
 void DEBUG_ShowMsg(const char *, ...) {}
 
 void DEBUG_HeavyWriteLogInstruction() {}
 
 #if C_DEBUG
-void LOG::operator()([[maybe_unused]] char const *buf, ...)
+void LOG::operator()([[maybe_unused]] const char* buf, ...)
 {
 	(void)d_type;     // Deliberately unused.
 	(void)d_severity; // Deliberately unused.
